from typing import Any, Callable, Dict
from unittest import TestCase

import pytest
from faker import Faker
from parametrize import parametrize
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import sqlalchemy_factories as factories
from faker_file_admin import app
from faker_file_admin.models import Upload

from ..registry import FILE_REGISTRY
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("SQLAlchemyIntegrationTestCase",)


STORAGE = FileSystemStorage(root_path="", rel_path="tmp")


class SQLAlchemyIntegrationTestCase(TestCase):
    """SQLAlchemy integration test case."""

    engine = create_engine("sqlite:///:memory:")
    Session = sessionmaker(bind=engine)
    session = Session()

    def setUp(self: "SQLAlchemyIntegrationTestCase"):
        Upload.metadata.create_all(self.engine)
        self.session.commit()

    def tearDown(self: "SQLAlchemyIntegrationTestCase"):
        Upload.metadata.drop_all(self.engine)
        FILE_REGISTRY.clean_up()

    FAKER: Faker

    @parametrize(
        "factory, kwargs",
        [
            (factories.UploadFactory, {}),
            (factories.UploadFactory, {"random_file": True}),
            (factories.UploadFactory, {"pdf_file": True}),
            (factories.UploadFactory, {"pptx_file": True}),
            (factories.UploadFactory, {"txt_file": True}),
            (factories.UploadFactory, {"zip_file": True}),
        ],
    )
    @pytest.mark.optional
    def test_file(
        self: "SQLAlchemyIntegrationTestCase",
        factory: Callable,
        kwargs: Dict[str, Any],
    ) -> None:
        """Test file."""
        with app.app_context():
            _upload = factory(**kwargs)
            self.session.commit()
            if kwargs:
                self.assertTrue(STORAGE.exists(_upload.file))



from typing import Callable, Optional, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import DEFAULT_FORMAT_FUNC, BytesValue, FileMixin, StringValue
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("GenericFileProvider",)


class GenericFileProvider(BaseProvider, FileMixin):
    """Generic file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.generic_file import GenericFileProvider

        FAKER = Faker()
        FAKER.add_provider(GenericFileProvider)

        file = FAKER.generic_file(
            content="<html><body><p>{{text}}</p></body></html>",
            extension="html",
        )

    Usage example with options:

    .. code-block:: python

        file = FAKER.generic_file(
            content="<html><body><p>{{text}}</p></body></html>",
            extension="html",
            prefix="zzz",
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.generic_file(
            content="<html><body><p>{{text}}</p></body></html>",
            extension="html",
            basename="index",
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
        )

    Usage example with AWS S3 storage:

    .. code-block:: python

        from faker_file.storages.aws_s3 import AWSS3Storage

        file = FAKER.generic_file(
            storage=AWSS3Storage(bucket_name="My-test-bucket"),
            content="<html><body><p>{{text}}</p></body></html>",
            extension="html",
        )
    """

    extension: str = None

    @overload
    def generic_file(
        self: "GenericFileProvider",
        content: Union[bytes, str],
        extension: str,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def generic_file(
        self: "GenericFileProvider",
        content: Union[bytes, str],
        extension: str,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def generic_file(
        self: "GenericFileProvider",
        content: Union[bytes, str],
        extension: str,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a generic file with given content.

        :param content: File content. If given, used as is.
        :param extension: File extension.
        :param storage: Storage class. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        self.extension = extension

        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        if self.generator is None:
            self.generator = Faker()

        data = {"content": content, "filename": filename, "storage": storage}

        if raw:
            if isinstance(content, bytes):
                raw_content = BytesValue(content)
            else:
                raw_content = BytesValue(
                    format_func(self.generator, content).encode()
                )
            raw_content.data = data
            return raw_content

        if isinstance(content, bytes):
            storage.write_bytes(filename, content)
        else:
            storage.write_text(
                filename,
                format_func(self.generator, content),
            )

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from io import BytesIO

from gtts import gTTS

from ...base.mp3_generator import BaseMp3Generator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("GttsMp3Generator",)


DEFAULT_LANG = "en"
DEFAULT_TLD = "com"


class GttsMp3Generator(BaseMp3Generator):
    """Google Text-to-Speech generator.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.mp3_file import Mp3FileProvider
        from faker_file.providers.mp3_file.generators.gtts_generator import (
            GttsMp3Generator,
        )

        FAKER = Faker()
        FAKER.add_provider(Mp3FileProvider)

        file = FAKER.mp3_file(
            mp3_generator_cls=GttsMp3Generator
        )
    """

    lang: str = DEFAULT_LANG
    tld: str = DEFAULT_TLD

    def handle_kwargs(self: "GttsMp3Generator", **kwargs) -> None:
        if "lang" in kwargs:
            self.lang = kwargs["lang"]
        if "tld" in kwargs:
            self.tld = kwargs["tld"]

    def generate(self: "GttsMp3Generator", **kwargs) -> bytes:
        """Generate MP3."""
        with BytesIO() as __fake_file:
            tts = gTTS(self.content, lang=self.lang, tld=self.tld)
            tts.write_to_fp(__fake_file)
            return __fake_file.getvalue()



import logging
import os
from typing import Optional

import paramiko

from .base import BaseStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("SFTPStorage",)

LOGGER = logging.getLogger(__name__)


class SFTPStorage(BaseStorage):
    """SFTP storage.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.txt_file import TxtFileProvider
        from faker_file.storages.sftp_storage import SFTPStorage

        FAKER = Faker()
        FAKER.add_provider(TxtFileProvider)

        # SFTP storage class
        STORAGE = SFTPStorage(
            host="0.0.0.0",
            username="foo",
            password="pass",
        )

        # Generate TXT file in the default directory
        txt_file = FAKER.txt_file(storage=STORAGE)

        # Another SFTP storage class, but inside a `/upload/another` directory
        STORAGE_SUB_DIR = SFTPStorage(
            host="0.0.0.0",
            username="foo",
            password="pass",
            root_path="/upload/another",
        )

        # Generate TXT file inside `/upload/another` directory
        txt_file = FAKER.txt_file(storage=STORAGE_SUB_DIR)
    """

    sftp: Optional[paramiko.SFTPClient] = None
    transport: Optional[paramiko.Transport] = None

    def __init__(
        self: "SFTPStorage",
        host: str,
        port: int = 22,
        username: str = "",
        password: Optional[str] = None,
        key: Optional[paramiko.PKey] = None,
        root_path: str = "",
        rel_path: str = "",
        *args,
        **kwargs,
    ) -> None:
        """
        :param host: The hostname of the SFTP server.
        :param port: The port of the SFTP server.
        :param username: The username to authenticate as.
        :param password: The password to use when connecting.
        :param key: The private key to use when connecting.
        :param root_path: Path of your files root directory.
        :param rel_path: Relative path (from root directory).
        """
        self.root_path = root_path
        self.rel_path = rel_path

        self.transport = paramiko.Transport((host, port))

        # Authentication
        try:
            if key:
                self.transport.connect(username=username, pkey=key)
            else:
                self.transport.connect(username=username, password=password)
            self.sftp = paramiko.SFTPClient.from_transport(self.transport)
        except Exception as e:
            LOGGER.exception(f"Failed to connect to SFTP server: {e}")
            raise
        super().__init__(*args, **kwargs)

    def _build_path(self: "SFTPStorage", filename: str) -> str:
        """Build the full path for a file."""
        return os.path.join(self.root_path, self.rel_path, filename)

    def generate_filename(
        self: "SFTPStorage",
        extension: str,
        prefix: Optional[str] = None,
        basename: Optional[str] = None,
    ) -> str:
        """Generate filename."""
        if not extension:
            LOGGER.error("File extension is required")
            raise ValueError("Extension shall be given!")

        if not basename:
            basename = self.generate_basename(prefix)

        return self._build_path(f"{basename}.{extension}")

        # if basename:
        #     return self._build_path(f"{basename}.{extension}")
        # else:
        #     with tempfile.NamedTemporaryFile(
        #         prefix=prefix,
        #         suffix=f".{extension}",
        #     ) as temp_file:
        #         return self._build_path(os.path.basename(temp_file.name))

    def write_text(
        self: "SFTPStorage",
        filename: str,
        data: str,
        encoding: Optional[str] = "utf-8",
    ) -> int:
        """Write text."""
        try:
            # Encode the text data into bytes before writing
            encoded_data = data.encode(encoding)
            with self.sftp.open(self._build_path(filename), "wb") as file:
                file.write(encoded_data)
                return 0
        except Exception as err:
            LOGGER.exception(f"Failed to write text to {filename}: {err}")
            return -1

    def write_bytes(self: "SFTPStorage", filename: str, data: bytes) -> int:
        """Write bytes."""
        try:
            with self.sftp.open(self._build_path(filename), "wb") as file:
                file.write(data)
                return 0
        except Exception as err:
            LOGGER.exception(f"Failed to write bytes to {filename}: {err}")
            return -1

    def exists(self: "SFTPStorage", filename: str) -> bool:
        """Check if file exists."""
        try:
            self.sftp.stat(self._build_path(filename))
            return True
        except IOError:
            return False

    def relpath(self: "SFTPStorage", filename: str) -> str:
        """Return relative path."""
        return os.path.relpath(self._build_path(filename), self.root_path)

    def abspath(self: "SFTPStorage", filename: str) -> str:
        """Return absolute path."""
        return self._build_path(filename)

    def unlink(self: "SFTPStorage", filename: str) -> None:
        """Remove a file."""
        try:
            self.sftp.remove(self._build_path(filename))
        except Exception as err:
            LOGGER.exception(f"Failed to remove {filename}: {err}")

    def close(self: "SFTPStorage"):
        """Explicitly close the connection."""
        if self.sftp:
            self.sftp.close()
        if self.transport:
            self.transport.close()

    def __del__(self: "SFTPStorage"):
        """Destructor to ensure connection is closed."""
        self.close()



from multiprocessing import Pool

from faker import Faker
from faker_file.providers.helpers.inner import (
    create_inner_docx_file,
    create_inner_epub_file,
    create_inner_pdf_file,
    create_inner_txt_file,
    fuzzy_choice_create_inner_file,
)
from faker_file.storages.filesystem import FileSystemStorage

FAKER = Faker()
STORAGE = FileSystemStorage()

# Document template
TEMPLATE = """
{{date}} {{city}}, {{country}}

Hello {{name}},

{{text}} {{text}} {{text}}

{{text}} {{text}} {{text}}

{{text}} {{text}} {{text}}

Address: {{address}}

Best regards,

{{name}}
{{address}}
{{phone_number}}
"""

kwargs = {"storage": STORAGE, "generator": FAKER, "content": TEMPLATE}

with Pool(processes=2) as pool:
    for _ in range(10):  # Number of times we want to run our function
        pool.apply_async(
            fuzzy_choice_create_inner_file,
            [
                [
                    (create_inner_docx_file, kwargs),
                    (create_inner_epub_file, kwargs),
                    (create_inner_pdf_file, kwargs),
                    (create_inner_txt_file, kwargs),
                ]
            ],
        )
    pool.close()
    pool.join()



from faker import Faker
from faker_file.providers.augment_file_from_dir import (
    AugmentFileFromDirProvider,
)
from faker_file.providers.docx_file import DocxFileProvider
from faker_file.providers.eml_file import EmlFileProvider
from faker_file.providers.odt_file import OdtFileProvider
from faker_file.providers.txt_file import TxtFileProvider

FAKER = Faker()
FAKER.add_provider(DocxFileProvider)
FAKER.add_provider(TxtFileProvider)
FAKER.add_provider(EmlFileProvider)
FAKER.add_provider(OdtFileProvider)
FAKER.add_provider(AugmentFileFromDirProvider)

# Create files to test `augment_file_from_dir` with
FAKER.docx_file()
FAKER.eml_file()
FAKER.odt_file()
FAKER.txt_file()

# We assume that directory "/tmp/tmp/" exists and contains
# files of `DOCX` and `ODT` formats.
augmented_file = FAKER.augment_file_from_dir(
    source_dir_path="/tmp/tmp/",
    extensions={"docx", "odt"},  # Pick only DOCX or ODT
)



# Required imports
from faker import Faker
from faker_file.providers.docx_file import DocxFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(DocxFileProvider)  # Register DocxFileProvider

# Generate DOCX file
docx_file = FAKER.docx_file()



from faker import Faker
from faker_file.providers.image.pil_generator import PilImageGenerator
from faker_file.providers.png_file import PngFileProvider

FAKER = Faker()
FAKER.add_provider(PngFileProvider)

png_file = FAKER.png_file(image_generator_cls=PilImageGenerator)



from typing import Union

from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("BaseMp3Generator",)


class BaseMp3Generator:
    """Base MP3 generator."""

    content: str
    generator: Union[Faker, Generator, Provider]

    def __init__(
        self: "BaseMp3Generator",
        content: str,
        generator: Union[Faker, Generator, Provider],
        **kwargs,
    ) -> None:
        """Constructor.

        :param content: Text content to generate MP3 from.
        :param generator: `Faker` instance that could contain useful
            information, such as active locale.
        :param kwargs: Dictionary with parameters (for MP3 generator
            specific tuning).
        """
        self.content = content
        self.generator = generator
        self.handle_kwargs(**kwargs)

    def handle_kwargs(self: "BaseMp3Generator", **kwargs):
        """Handle kwargs."""

    def generate(self: "BaseMp3Generator", **kwargs) -> bytes:
        raise NotImplementedError("Method `generate` is not implemented.")
